from __future__ import annotations
from abc import ABCMeta, abstractmethod
from typing import Union

from aiostorage_orm.operation_result import OperationResult


class AIOStorageItem(metaclass=ABCMeta):
    """
        Базовая модель для объекта БД
        - Для создания модели на основе текущей, необходимо определить класс
          конфигурации Meta и создать поля объекта, например:

            class MyModel(AIOStorageItem):
                date_time: float
                any_value: int

                class Meta:
                    table = "subsystem.{subsystem_id}.tag.{tag_id}"
                    ttl = 3600  # sec, default None
    """

    @classmethod
    @abstractmethod
    async def get(cls, _item, **kwargs) -> Union[AIOStorageItem, None]:
        """
            Получение одного объекта по выбранному фильтру

                AIOStorageItem.get(subsystem_id=10, tag_id=55)
                AIOStorageItem.get(_item=AIOStorageItem(subsystem_id=10))
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def filter(cls, _items, **kwargs) -> list:
        """
            Получение объектов по фильтру переданных аргументов, например:

                AIOStorageItem.filter(subsystem_id=10, tag_id=55)
                AIOStorageItem.filter(_items=[AIOStorageItem(subsystem_id=10), ...])
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def using(cls, db_instance) -> AIOStorageItem:
        """
            Выполнение операций с БД путём direct-указания используемого
            подключения, например:

                another_client: redis.Redis = redis.Redis(host="8.8.8.8", db=12)
                AIOStorageItem.using(db_instance=another_client).get(subsystem_id=10)
        """
        raise NotImplementedError

    @abstractmethod
    async def save(self) -> OperationResult:
        """
            Одиночная вставка
        """
        raise NotImplementedError

    @abstractmethod
    async def delete(self) -> OperationResult:
        """
            Удаление одного элемента
        """
        raise NotImplementedError

    @abstractmethod
    def set_ttl(self, new_ttl) -> None:
        """ Установка настройки времени жизни объекта 'на лету' """
        raise NotImplementedError

    @abstractmethod
    async def set_frame_size(self, new_frame_size) -> None:
        """ Установка настройки максимального размера frame'а 'на лету' """
        raise NotImplementedError



import logging
import time
from typing import Union
from typing import TypeVar

import redis.asyncio as redis
from redis.asyncio.client import Pipeline

from .aioredis_frame import AIORedisFrame
from .aioredis_item import AIORedisItem
from .aioredis_item import T as SubclassItemType
from ..operation_result import OperationResult
from ..operation_result import OperationStatus

from ..aiostorage_orm import AIOStorageORM

ChildItem = TypeVar('ChildItem', bound=AIORedisItem)


class AIORedisORM(AIOStorageORM):
    """ Работа с БД Redis через объектное представление """
    _pipe: Pipeline
    _client: redis.Redis
    _frame: AIORedisFrame

    def __init__(
        self,
        client: Union[redis.Redis, None] = None,
        host: Union[str, None] = None,
        port: int = 6379,
        db: int = 0,
    ) -> None:
        if client:
            self._client = client
        elif host:
            self._client = redis.Redis(host=host, port=port, db=db)
        else:
            raise Exception("AIOStorageORM-init must contains redis_client or host values...")

        self._pipe = self._client.pipeline()
        if not AIORedisItem._db_instance:
            AIORedisItem._set_global_instance(db_instance=self._client)

        self._frame = AIORedisFrame(client=self._client)

    async def init(self) -> None:
        """
        Проверка подключения к redis
        """
        await self._raise_for_connection()

    async def _raise_for_connection(self):
        """
        Retry проверка подключения к redis

        Ошибки:
            При отсутствии подключения к redis выбрасывается исключение ConnectionError
        """
        time_for_retry = [5, 2, 1, 0.5, 0.1]
        while time_for_retry:
            if await AIORedisItem._is_connected(AIORedisItem._db_instance):  # type: ignore
                return
            time.sleep(time_for_retry.pop())
        raise ConnectionError("Redis connection error...")

    @property
    def frame(self) -> AIORedisFrame:
        """ Подготовленный frame для работы со списками значений """
        return self._frame

    async def save(self, item: AIORedisItem) -> OperationResult:
        """ Одиночная вставка """
        return await item.save()

    async def bulk_create(self, items: list[SubclassItemType]) -> OperationResult:
        """ Групповая вставка """
        try:
            if hasattr(items[0], "_ttl") and items[0]._ttl:
                for redis_item in items:
                    for key, value in redis_item.mapping.items():
                        self._pipe.set(name=key, value=value, ex=redis_item._ttl)
            else:
                for redis_item in items:
                    self._pipe.mset(mapping=redis_item.mapping)
            await self._pipe.execute()
            return OperationResult(status=OperationStatus.success)
        except Exception as exception:
            self._on_error_actions(exception=exception)
            return OperationResult(
                status=OperationStatus.failed,
                message=str(exception),
            )

    async def bulk_delete(self, items: list[ChildItem]) -> OperationResult:
        """
            Удаление списка элементов
        """
        try:
            for redis_item in items:
                self._pipe.delete(*[key for key in redis_item.mapping.keys()])
            await self._pipe.execute()
            return OperationResult(status=OperationStatus.success)
        except Exception as exception:
            self._on_error_actions(exception=exception)
            return OperationResult(
                status=OperationStatus.failed,
                message=str(exception),
            )

    async def delete(self, item: AIORedisItem) -> OperationResult:
        """
            Удаление одного элемента
        """
        return await item.delete()

    def _on_error_actions(self, exception: Exception) -> None:
        """
            Действия, выполняющиеся в случае возникновения исключения
                во время вставки, сохранения, получения данных из БД
        """
        logging.exception(exception)



from abc import ABCMeta, abstractmethod
from typing import Any

from .operation_result import OperationResult


class AIOStorageFrame(metaclass=ABCMeta):
    _db_instance: Any

    @abstractmethod
    async def add(self, item_or_items) -> OperationResult:
        raise NotImplementedError

    @abstractmethod
    async def bulk_create(self, items) -> OperationResult:
        raise NotImplementedError

    @abstractmethod
    async def clear(self, item) -> OperationResult:
        raise NotImplementedError

    @abstractmethod
    async def get(self, item, start_index: int = 0, end_index: int = -1) -> list:
        raise NotImplementedError



from .redis_impl import AIORedisORM
from .redis_impl import AIORedisItem
from .redis_impl import AIORedisFrame

from .aiostorage_orm import AIOStorageORM
from .aiostorage_item import AIOStorageItem

from .operation_result import OperationResult
from .operation_result import OperationStatus

from .exceptions import NotEnoughParamsException
from .exceptions import MultipleGetParamsException
from .exceptions import OrmNotInitializedException



import pickle
import logging
from typing import Any
from typing import Union
import redis.asyncio as redis
from redis.asyncio.client import Pipeline
from redis.commands.core import AsyncScript

from .aioredis_item import AIORedisItem
from .aioredis_item import T as SubclassItemType
from ..operation_result import OperationResult
from ..operation_result import OperationStatus
from ..aiostorage_frame import AIOStorageFrame


class AIORedisFrame(AIOStorageFrame):
    """
    Работа с искусственно ограниченным frame'ом объектов

    Frame - массив сериализованных объектов определенного размера
    Размер frame'а задается в объекте типа AIOStorageORM, в блоке Meta -> frame_size
    Хранимый скрипт - lua-скрипт, который добавляется в Redis путем
                      вызова register_script. Он позволяет выполнять
                      установленную последовательность действий без
                      переключения контекста (python-redis), а также
                      все операции выполняются "за один раз", что
                      гарантирует атомарность всего процесса

    """
    DEFAULT_QUEUE_SIZE: int = 100
    # Префикс записей frame'ов. Нужен для исключения потенциальных пересечений
    FRAME_PREFIX: str = "frame."
    QUEUE_START_INDEX: int = 0
    QUEUE_END_INDEX: int = -1

    _pipe: Pipeline
    _client: redis.Redis
    _queue_add: AsyncScript  # Хранимый скрипт Redis

    def __init__(self, client: redis.Redis) -> None:
        self._client = client
        self._pipe = client.pipeline()
        # Инициализация хранимого скрипта в Redis для атомарного добавления
        #   в конец одного элемента и удаления первого элемента, если размер
        #   списка достиг лимита
        # Сигнатура вызова после регистрации:
        #   self._queue_add(keys=["example_key"], args=["my_value", 10], client=...)
        #   , где 10 - это размер списка (лимит)
        #         client - это клиет, в котором будет вызван скрипт (им может быть
        #                  pipeline)
        self._queue_add = self._client.register_script("""
            redis.call('rpush', KEYS[1], ARGV[1])
            if redis.call('llen', KEYS[1]) > tonumber(ARGV[2]) then
                redis.call('lpop', KEYS[1])
            end
        """)
        # Подрезка списков, согласно установленной в subclass'е величине
        AIORedisItem._frame_ltrim = self.ltrim_by_item

    async def add(
        self,
        item_or_items: Union[SubclassItemType, list[SubclassItemType]],
    ) -> OperationResult:
        """
        Добавление объект(а/ов) во frame

        В списке объектов приемлемо использовать разнородные данные
            (т.е. переданные объекты могут не принадлежать одному классу)

        """
        try:
            if isinstance(item_or_items, AIORedisItem):
                await self._add_item(item=item_or_items)
            else:
                [await self._add_item(item) for item in item_or_items]  # type: ignore
        except Exception as exception:
            await self._pipe.reset()
            logging.exception(exception)
            return OperationResult(
                status=OperationStatus.failed,
                message=str(exception),
            )
        await self._pipe.execute()
        return OperationResult(status=OperationStatus.success)

    def _get_frame_size(self, item: AIORedisItem) -> int:
        if hasattr(item, "_frame_size") and item._frame_size:
            queue_size = item._frame_size
        else:
            queue_size = self.DEFAULT_QUEUE_SIZE
        return queue_size

    async def _add_item(self, item: AIORedisItem) -> None:
        """
        Добавление объекта в БД

        - сериализация данных объекта
        - вызов хранимого скрипта Redis для атомарной операции
          наполнения ограниченного списка

        """
        # item._table содержит строку с подставленными параметрами текущего объекта
        key: str = self._make_key(item=item)
        object_key: bytes = key.encode()
        values: tuple = tuple(
            item.__dict__[key]
            for key in sorted(item.__annotations__.keys())
            if not key.startswith("_")
        )
        serialized_object: bytes = pickle.dumps(values)
        queue_size: int = self._get_frame_size(item=item)
        # Вызов запускает хранимый скрипт, который нуждается в
        #   - KEYS[1] (keys[0]) - ключ объекта Redis
        #   - ARGV[1] (args[0]) - данные для хранения
        #   - ARGV[2] (args[1]) - лимит размера списка
        await self._queue_add(
            keys=[object_key],
            args=[serialized_object, queue_size],
            client=self._pipe,
        )
        # _pipe может быть использован в цикле для списка объектов, поэтому
        #   осознанно не выполняем (execute)

    async def bulk_create(self, items: list[SubclassItemType]) -> OperationResult:
        return await self.add(item_or_items=items)

    async def clear(self, item: AIORedisItem) -> OperationResult:
        """ Удаление frame'а из БД """
        try:
            key: str = self._make_key(item=item)
            await self._client.delete(key)
        except Exception as exception:
            logging.exception(exception)
            return OperationResult(
                status=OperationStatus.failed,
                message=str(exception),
            )
        return OperationResult(status=OperationStatus.success)

    def _values_to_items(
        self,
        values: list[tuple],
        item: SubclassItemType,
    ) -> list[SubclassItemType]:
        """ Десериализация и создание объектов типа item.__class__ со списком значений """
        T: type = item.__class__  # Класс десериализуемого объекта, нужен для вызова конструктора
        # Получение параметров класса (subsytem, tag_id и т.п.)
        params: dict[str, Any] = {key: item.__dict__[key] for key in item._keys_positions.keys()}
        # Сортировка имён для правильного расположения значений в атрибуты
        attr_names: list[str] = list(sorted(T.__annotations__.keys()))
        # Формирование словарей для инициализации объектов
        init_dicts: list[dict] = [dict(zip(attr_names, item_values)) for item_values in values]
        result: list[SubclassItemType] = [T(**(init_dict | params)) for init_dict in init_dicts]
        return result

    def _make_key(self, item: AIORedisItem) -> str:
        return f"{self.FRAME_PREFIX}{item._table}"

    async def ltrim_by_item(self, item: AIORedisItem) -> None:
        """ Форматирование(обрезка) длины очереди в соответствии с frame_size Item'а """
        key: str = self._make_key(item=item)
        total_items_count: int = await self._client.llen(key)
        queue_size: int = self._get_frame_size(item=item)
        if total_items_count > queue_size:
            await self._client.ltrim(key, -queue_size, self.QUEUE_END_INDEX)

    async def get(
        self,
        item: SubclassItemType,
        start_index: int = QUEUE_START_INDEX,
        end_index: int = QUEUE_END_INDEX,
    ) -> list[SubclassItemType]:
        """
        Получение данных из БД и приведение их к соответствующим типам объектов

        item: AIORedisItem - объект с подготовленными для поиска параметрами
        start_index: int - индекс начального элемента frame'а (отсчет начинается с 0,)
        end_index: int - индекс последнего элемента (включительно)

        * start_index - самый старый объект, end_index - самый свежий объект

        """
        key: str = self._make_key(item=item)
        serialized_values: list[bytes] = await self._client.lrange(key, start_index, end_index)
        # Десериализация полученных данных
        all_values_list: list[tuple] = [pickle.loads(v) for v in serialized_values]
        # Формирование объектов из полученных данных
        items: list[SubclassItemType] = self._values_to_items(
            values=all_values_list,
            item=item,
        )
        return items



from __future__ import annotations
import re
import copy
import uuid
import pickle
from contextlib import suppress

import redis.asyncio as redis
from redis.exceptions import ConnectionError
import itertools
from typing import (
    Any,
    cast,
    Union,
    Optional,
    Mapping,
    Type,
    TypeVar,
    Coroutine,
    Callable
)

from ..aiostorage_item import AIOStorageItem
from ..operation_result import OperationResult
from ..operation_result import OperationStatus
from ..exceptions import MultipleGetParamsException
from ..exceptions import NotEnoughParamsException
from ..exceptions import OrmNotInitializedException

# Redis: переопределения типов для корректной работы линтеров
_Value = Union[bytes, float, int, str]
_Key = Union[str, bytes]
ResponseT = Any

T = TypeVar('T', bound='AIORedisItem')
IN_SUFFIX = "__in"
KEYS_DELIMITER = "."


class AIORedisItem(AIOStorageItem):
    _table: str
    _keys_positions: dict[str, int]
    _params: Mapping[_Key, _Value]
    _db_instance: Union[redis.Redis, None] = None
    _frame_ltrim: Optional[Callable[[AIORedisItem], Coroutine[Any, Any, None]]] = None
    _frame_size: int = 0
    _ttl: Optional[int] = None

    class Meta:
        table: str = ""  # Pattern имени записи, например, "subsystem.{subsystem_id}.tag.{tag_id}"
        ttl: Optional[int] = None  # Время жизни объекта в базе данных
        frame_size: Optional[int] = 100  # Максимальный размер frame'а

    def __init_subclass__(cls) -> None:
        cls._keys_positions = {
            index.replace("{", "").replace("}", ""): key
            for key, index in enumerate(cls.Meta.table.split(KEYS_DELIMITER))
            if index.startswith("{") and index.endswith("}")
        }
        for param in cls._keys_positions.keys():
            if param in cls.__annotations__:
                del cls.__annotations__[param]
        # Аргумент, который используется для дальнейшей проверки и работы
        if hasattr(cls.Meta, "frame_size"):
            setattr(cls, "_frame_size", cls.Meta.frame_size)

    @classmethod
    def _make_kwargs_from_objects(cls: Type[T], objects: list[T]) -> dict:
        """
            Конкатенация атрибутов объектов и их подготовка для
                использования в качестве фильтров
        """
        result_kwargs: dict = {}
        for obj in objects:
            for key, position in obj._keys_positions.items():
                value: str = obj._table.split(KEYS_DELIMITER)[position]
                if key in result_kwargs:
                    result_kwargs[key] += str(value)
                else:
                    result_kwargs[key] = str(value)

        for key in result_kwargs.keys():
            result_kwargs[key] = f"[{result_kwargs[key]}]"

        return result_kwargs

    def set_ttl(self, new_ttl: int) -> None:
        """ Установка настройки времени жизни объекта 'на лету' """
        setattr(self, "_ttl", new_ttl)

    async def set_frame_size(self, new_frame_size: int = 0) -> None:
        """ Установка настройки максимального размера frame'а 'на лету' """
        meta_frame_size: int | None = self.Meta.frame_size if hasattr(self.Meta, "frame_size") else None
        old_frame_size: int = self._frame_size or meta_frame_size or 0
        # При одинаковых значениях ничего не делать
        if old_frame_size == new_frame_size:
            return
        # Если передан пустой аргумент, нужно установить значение по умолчанию
        if not new_frame_size and self.Meta.frame_size:
            new_frame_size = self.Meta.frame_size
        # Аргумент, который используется для дальнейшей проверки и работы
        setattr(self, "_frame_size", new_frame_size)

        # Проверка необходимости подрезки frame'а
        if old_frame_size > new_frame_size:
            # Подрезка фрейма в БД
            if self._frame_ltrim:
                await self._frame_ltrim(self)

    async def init_frame(self) -> None:
        """
        Инициализация для выполнения подрезки фрейма
        """
        if self._frame_ltrim:
            await self._frame_ltrim(self)
        else:
            raise OrmNotInitializedException("ORM not initialized")

    def __init__(self, **kwargs) -> None:
        # Установка атрибутов из конструктора
        for config_key in ("ttl", "frame_size"):
            if config_key in kwargs.keys():
                setattr(self, f"_{config_key}", kwargs[config_key])
                del kwargs[config_key]
        # Формирование полей модели из переданных дочернему классу аргументов
        [self.__dict__.__setitem__(key, value) for key, value in kwargs.items()]  # type: ignore
        # Формирование изолированной среды с данными класса для дальнейшей работы с БД
        self._table = self.__class__.Meta.table.format(**kwargs)
        self._params = {
            key: kwargs.get(key, None)
            for key in self.__class__.__annotations__
        }
        # Перегрузка методов для экземпляра класса
        self.using = self.instance_using  # type: ignore

    def __getattr__(self, attr_name: str):
        return object.__getattribute__(self, attr_name)

    @classmethod
    def _set_global_instance(cls: Type[T], db_instance: redis.Redis) -> None:
        """ Установка глобальной ссылки на БД во время первого подключения """
        cls._db_instance = db_instance

    @classmethod
    def _get_keys_list(cls: Type[T], prefix: str) -> list[bytes]:
        """ Формирование ключей для поиска в БД на основе префикса и атрибутов класса"""
        return [f"{prefix}.{key}".encode() for key in cls.__annotations__.keys()]

    @staticmethod
    async def _is_connected(db_instance: redis.Redis) -> bool:
        """ Проверка наличия подключения к серверу """
        try:
            await db_instance.ping()  # type: ignore
        except ConnectionError:
            return False

        return True

    @classmethod
    async def get(cls: Type[T], _item: Union[T, None] = None, **kwargs) -> Union[T, None]:
        """
            Получение одного объекта по выбранному фильтру

                AIOStorageItem.get(subsystem_id=10, tag_id=55)
                AIOStorageItem.get(_item=AIOStorageItem(subsystem_id=10))
        """
        if not cls._db_instance or not await cls._is_connected(db_instance=cls._db_instance):
            raise Exception("Redis database not connected...")
        if len(kwargs) and _item:
            raise Exception(f"{cls.__name__}.get() has _item and kwargs. It's not possible.")
        filter: str
        if _item:
            filter = _item._table
        else:
            # Разобрать готовым методом аргуметы в список фильтров
            filters_list: list[str] = cls._get_filters_by_kwargs(**kwargs)
            if len(filters_list) > 1:
                raise MultipleGetParamsException(
                    f"{cls.__name__} invalid (uses __in) params to get method..."
                )
            filter = filters_list[0]
            # Использование маски для выборки одного объекта не предусмотрено
            if not filter or "*" in filter:
                raise NotEnoughParamsException(
                    f"{cls.__name__} not enough params to get method..."
                )
        keys: list[bytes] = cls._get_keys_list(prefix=filter)
        values: list[bytes] = cast(list[bytes], await cls._db_instance.mget(keys))
        if not [v for v in values if v]:
            return None
        finded_objects: list[T] = cls._objects_from_db_items(items=dict(zip(keys, values)))
        result: Union[T, None] = finded_objects[0]
        return result

    @classmethod
    async def filter(cls: Type[T], _items: Union[list[T], None] = None, **kwargs) -> list[T]:
        """
            Получение объектов по фильтру переданных аргументов, например:

                AIOStorageItem.filter(subsystem_id=10, tag_id=55)
                AIOStorageItem.filter(_items=[AIOStorageItem(subsystem_id=10), ...])
        """
        if not cls._db_instance or not await cls._is_connected(db_instance=cls._db_instance):
            raise Exception("Redis database not connected...")
        if not len(kwargs) and not _items:
            raise Exception(f"{cls.__name__}.filter() has empty filter. OOM possible.")
        if len(kwargs) and _items:
            raise Exception(f"{cls.__name__}.filter() has _items and kwargs. It's not possible.")
        filters_list: list[str]
        keys: list[bytes] = []
        if _items:
            filters_list = [item._table for item in _items]
            for filter in filters_list:
                keys += cls._get_keys_list(prefix=filter)
        else:
            # Формирование списка фильтров для возможности поиска входящих в список
            filters_list = cls._get_filters_by_kwargs(**kwargs)
            for filter in filters_list:
                keys_list: list[bytes] = cls._get_keys_list(prefix=filter)
                if [key for key in keys_list if "*" in str(key)]:
                    # Если не передан один из параметров и нужен поиск по ключам
                    keys += await cls._db_instance.keys(pattern=filter + ".*")
                else:
                    # Если все параметры присутствуют, то можно использовать только
                    #   имена атрибутов
                    keys += keys_list

        values: list[bytes] = cast(list[bytes], await cls._db_instance.mget(keys))
        # Очистка пустых значений полученных данных
        if not [v for v in values if v]:
            return []

        result: list[T] = cls._objects_from_db_items(items=dict(zip(keys, values)))

        return result

    @classmethod
    def _all_fields_is_empty(cls: Type[T], items: dict[bytes, bytes], fields: list[bytes]) -> bool:
        """ Проверка на отсутствие всех значений создаваемого объекта """
        for field in fields:
            with suppress(pickle.UnpicklingError):
                if field in items and items[field] and pickle.loads(items[field]) is not None:
                    return False
        return True

    @classmethod
    def _get_src_values_for_meta(cls: Type[T], table: str) -> dict:
        """ Получение значений данных Meta класса"""
        src_values_for_meta: dict = dict()
        src_values: list[str] = table.split(".")
        for key, position in cls._keys_positions.items():
            if len(src_values[position]) == 36:  # UUID
                src_values_for_meta[key] = uuid.UUID(src_values[position])
            else:  # int
                src_values_for_meta[key] = int(src_values[position])
        return src_values_for_meta

    @classmethod
    def _objects_from_db_items(cls: Type[T], items: dict[bytes, bytes]) -> list[T]:
        """ Формирование cls(RedisItem)-объектов из данных базы """
        class_fields: dict = {}
        for redis_key, src_value in items.items():
            # Определение имени таблицы (без имени поля)
            table = redis_key.decode().rsplit(KEYS_DELIMITER, 1)[0]
            # Определение имени поля
            field_name = redis_key.decode().rsplit(KEYS_DELIMITER, 1)[1]

            if cls._all_fields_is_empty(items=items, fields=[redis_key]):
                continue

            # Подготовка базовых данных для формирования объекта
            if not class_fields.get(table):
                class_fields[table] = {
                    field_name: pickle.loads(src_value)
                }
            else:
                class_fields[table] |= {
                    field_name: pickle.loads(src_value)
                }

        result_items = []
        for table, class_args in class_fields.items():
            src_values_for_meta: dict = cls._get_src_values_for_meta(table=table)

            result_items.append(
                cls(**(class_args | src_values_for_meta))
            )

        return result_items

    @staticmethod
    def _get_list_of_prepared_kwargs(**kwargs: dict) -> list[dict]:
        """
            Подготовка списка фильтров из словарей:
                - исходный словарь разделить:
                    - базовый (без списков в значениях)
                    - расширенный (со списками в значениях)
                - получить множество комбинаций расширенного словаря
                - скомбинировать

            Examples:
                >>>
                kwargs = {"param1__in": [1, 2], "param2__in": [3, 4]}
                result = [
                    {"param1": 1, "param2": 3},
                    {"param1": 1, "param2": 4},
                    {"param1": 2, "param2": 3},
                    {"param1": 2, "param2": 4},
                ]
        """
        basic_kwargs: dict = {}
        extend_kwargs: dict = {}
        # Разделение на словари "с" и "без" списков в значениях
        for key, value in kwargs.items():
            if not key.endswith(IN_SUFFIX):
                basic_kwargs[key] = value
            else:
                extend_kwargs[key.strip(IN_SUFFIX)] = value
        # Формирование итоговых словарей
        result_kwargs: list[dict] = []
        if extend_kwargs:
            # Получить множество комбинаций расширенного словаря
            mixed_kwargs: list[dict] = list(
                dict(zip(extend_kwargs.keys(), values))
                for values in itertools.product(*extend_kwargs.values())
            )
            # Обогатить расширенные словари базовым
            result_kwargs = [mixed_item | basic_kwargs for mixed_item in mixed_kwargs]
        else:
            result_kwargs = [basic_kwargs]

        return result_kwargs

    @classmethod
    def _get_filters_by_kwargs(cls: Type[T], **kwargs: dict) -> list[str]:
        """ Подготовка списка паттернов поиска """
        table: str = cls.Meta.table
        # Шаблон для поиска аргументов, которе не были переданы
        patterns: list[str] = re.findall(r'\{[^\}]*\}', table)
        str_filters: list[str] = []
        # Получение сырого списка фильтров
        prepared_kwargs_list: list[dict] = cls._get_list_of_prepared_kwargs(**kwargs)
        # Замена аргументов, которые не переданы, на звездочку
        for prepared_kwargs in prepared_kwargs_list:
            for pattern in patterns:
                clean_key: str = pattern.strip("{").strip("}")
                if clean_key not in prepared_kwargs:
                    table = table.replace(pattern, "*")
            # Заполнение паттерна поиска
            str_filters.append(table.format(**prepared_kwargs))

        return str_filters

    @property
    def mapping(self) -> Mapping[_Key, _Value]:
        """ Формирование ключей и значений для БД """
        return {
            KEYS_DELIMITER.join([self._table, str(key)]): pickle.dumps(value)
            for key, value in self._params.items()
        }

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self._table=}, "
            f"{self._keys_positions=}, {self._params=})"
        )

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            return self._params == other._params and self._table == other._table

        return False

    def instance_using(self: T, db_instance: Union[redis.Redis, None] = None) -> T:
        """
            Выполнение операций с БД путём direct-указания используемого
            подключения, например:

                another_client: redis.Redis = redis.Redis(host="8.8.8.8", db=12)
                storage_item_instance.using(db_instance=another_client).save()

            Создаётся копия объекта для работы через "неглобальное" подключение к Redis
        """
        copied_instance: T = copy.copy(self)
        copied_instance._db_instance = db_instance
        return copied_instance

    @classmethod
    def using(cls: Type[T], db_instance: Union[redis.Redis, None] = None) -> T:
        """
            Выполнение операций с БД путём direct-указания используемого
            подключения, например:

                another_client: redis.Redis = redis.Redis(host="8.8.8.8", db=12)
                AIOStorageItem.using(db_instance=another_client).get(subsystem_id=10)

            Создаётся копия класса для работы через "неглобальное" подключение к Redis
        """
        class CopiedClass(cls):  # type: ignore
            _db_instance = db_instance
        CopiedClass.__annotations__.update(cls.__annotations__)
        CopiedClass.__name__ = cls.__name__
        return cast(T, CopiedClass)

    async def save(self) -> OperationResult:
        """ Одиночная вставка """
        try:
            if not self._db_instance or not await self._is_connected(db_instance=self._db_instance):
                raise Exception("Redis database not connected...")
            for key, value in self.mapping.items():
                expiration: Union[int, None] = self._ttl if hasattr(self, "_ttl") else None
                await self._db_instance.set(name=key, value=value, ex=expiration)
            return OperationResult(status=OperationStatus.success)
        except Exception as exception:
            return OperationResult(
                status=OperationStatus.failed,
                message=str(exception),
            )

    async def delete(self) -> OperationResult:
        """ Удаление одного элемента """
        try:
            if not self._db_instance or not await self._is_connected(db_instance=self._db_instance):
                raise Exception("Redis database not connected...")
            await self._db_instance.delete(*[key for key in self.mapping.keys()])
            return OperationResult(status=OperationStatus.success)
        except Exception as exception:
            return OperationResult(
                status=OperationStatus.failed,
                message=str(exception),
            )



from abc import ABCMeta, abstractmethod, abstractproperty
from typing import Any, Union

from .aiostorage_frame import AIOStorageFrame
from .operation_result import OperationResult


class AIOStorageORM(metaclass=ABCMeta):
    _db_instance: Any

    @abstractmethod
    def __init__(
        self,
        client: Any = None,
        host: Union[str, None] = None,
        port: int = 6379,
        db: int = 0,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    async def init(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def save(self, item) -> OperationResult:
        raise NotImplementedError

    @abstractmethod
    async def bulk_create(self, items: list) -> OperationResult:
        raise NotImplementedError

    @abstractmethod
    async def bulk_delete(self, items: list) -> OperationResult:
        raise NotImplementedError

    @abstractmethod
    async def delete(self, item) -> OperationResult:
        raise NotImplementedError

    @abstractproperty
    def frame(self) -> AIOStorageFrame:
        raise NotImplementedError



from enum import Enum
from typing import Union


class OperationStatus(Enum):
    """ Статус операции записи/чтения из БД """
    success = True
    failed = False


class OperationResult:
    """ Результат записи/чтения из БД """
    status: OperationStatus
    message: str

    def __init__(self, status: Union[OperationStatus, bool], message: str = "") -> None:
        self.status = OperationStatus(status)
        self.message = message

    @property
    def ok(self) -> bool:
        return self.status == OperationStatus.success

    def __str__(self) -> str:
        message: str = f", message={self.message}" if self.message else ""
        return f"{self.__class__.__name__}: status={self.status}{message}"

    def __repr__(self) -> str:
        message: str = f", message={self.message}" if self.message else ""
        return f"{self.__class__.__name__}: status={self.status}{message}"
